#include "server.h"

#include <stdlib.h>

#include <wlr/backend.h>
#include <wlr/types/wlr_compositor.h>
#include <wlr/types/wlr_data_device.h>
#include <wlr/types/wlr_input_device.h>
#include <wlr/types/wlr_xdg_shell.h>
#include <wlr/util/log.h>

#include "input/keyboard.h"
#include "input/seat.h"
#include "output.h"

/* defines handle_$member_destroy for destroy signals of half_server's members
 * destroys the server when the member's destroyed
 */
#define TAKE_SERVER_WITH(member)                                               \
	static void handle_##member##_destroy(                                       \
			struct wl_listener* listener, void* data) {                              \
		struct half_server* half_server =                                          \
				wl_container_of(listener, half_server, member##_destroy);              \
		(void)data;                                                                \
		/* prevent double free */                                                  \
		half_server->member = NULL;                                                \
		half_server_destroy(half_server);                                          \
	}

// TODO: attempt to recreate these before killing the server
TAKE_SERVER_WITH(wl_display)
TAKE_SERVER_WITH(wlr_backend)
TAKE_SERVER_WITH(wlr_compositor)
TAKE_SERVER_WITH(wlr_xdg_shell)
TAKE_SERVER_WITH(wlr_data_device_manager)

static void handle_wlr_backend_new_input(struct wl_listener*, void* data);
static void handle_wlr_backend_new_output(struct wl_listener*, void* data);

struct half_server* half_server_create() {
	struct half_server* half_server = calloc(1, sizeof(struct half_server));
	if(!half_server) {
		fprintf(stderr, "Failed to allocate half_server\n");
		return NULL;
	}

	// set up server members and their event handlers

	// wl_display
	half_server->wl_display = wl_display_create();
	if(!half_server->wl_display) {
		fprintf(stderr, "Failed to create wl_display\n");
		half_server_destroy(half_server);
		return NULL;
	}
	half_server->wl_display_destroy.notify = handle_wl_display_destroy;
	wl_display_add_destroy_listener(
			half_server->wl_display, &half_server->wl_display_destroy);

	// wlr_backend
	half_server->wlr_backend =
			wlr_backend_autocreate(half_server->wl_display, NULL);
	if(!half_server->wlr_backend) {
		fprintf(stderr, "Failed to create wlr_backend\n");
		half_server_destroy(half_server);
		return NULL;
	}
	half_server->wlr_backend_destroy.notify = handle_wlr_backend_destroy;
	wl_signal_add(
			&half_server->wlr_backend->events.destroy,
			&half_server->wlr_backend_destroy);
	half_server->wlr_backend_new_input.notify = handle_wlr_backend_new_input;
	wl_signal_add(
			&half_server->wlr_backend->events.new_input,
			&half_server->wlr_backend_new_input);
	half_server->wlr_backend_new_output.notify = handle_wlr_backend_new_output;
	wl_signal_add(
			&half_server->wlr_backend->events.new_output,
			&half_server->wlr_backend_new_output);

	// wlr_compositor
	half_server->wlr_compositor = wlr_compositor_create(
			half_server->wl_display,
			wlr_backend_get_renderer(half_server->wlr_backend));
	if(!half_server->wlr_compositor) {
		fprintf(stderr, "Failed to create wlr_compositor\n");
		half_server_destroy(half_server);
		return NULL;
	}
	half_server->wlr_compositor_destroy.notify = handle_wlr_compositor_destroy;
	wl_signal_add(
			&half_server->wlr_compositor->events.destroy,
			&half_server->wlr_backend_destroy);

	// wlr_xdg_shell
	half_server->wlr_xdg_shell = wlr_xdg_shell_create(half_server->wl_display);
	if(!half_server->wlr_xdg_shell) {
		fprintf(stderr, "Failed to create wlr_xdg_shell\n");
		half_server_destroy(half_server);
		return NULL;
	}
	half_server->wlr_xdg_shell_destroy.notify = handle_wlr_xdg_shell_destroy;
	wl_signal_add(
			&half_server->wlr_xdg_shell->events.destroy,
			&half_server->wlr_xdg_shell_destroy);

	// wlr_data_device_manager
	half_server->wlr_data_device_manager =
			wlr_data_device_manager_create(half_server->wl_display);
	if(!half_server->wlr_data_device_manager) {
		fprintf(stderr, "Failed to create wlr_data_device_manager\n");
		half_server_destroy(half_server);
		return NULL;
	}
	half_server->wlr_data_device_manager_destroy.notify =
			handle_wlr_data_device_manager_destroy;
	wl_signal_add(
			&half_server->wlr_data_device_manager->events.destroy,
			&half_server->wlr_data_device_manager_destroy);

	wl_list_init(&half_server->seats);
	wl_list_init(&half_server->outputs);

	// add a primary seat - there should always be at least one in the seats list
	struct half_seat* primary_seat =
			half_seat_create(half_server->wl_display, "seat0");
	wl_list_insert(&half_server->seats, &primary_seat->link);

	// shm
	if(wl_display_init_shm(half_server->wl_display) < 0) {
		fprintf(stderr, "Failed to init shm\n");
		half_server_destroy(half_server);
		return NULL;
	}

	// socket
	half_server->socket = wl_display_add_socket_auto(half_server->wl_display);
	if(!half_server->socket) {
		fprintf(stderr, "Failed to create wayland socket\n");
		half_server_destroy(half_server);
		return NULL;
	}

	return half_server;
}

void half_server_destroy(struct half_server* half_server) {
	if(!half_server) {
		return;
	}

	// unsubscribe from everyone's all events (inits are for safety, see WL docs)
	wl_list_init(&half_server->wl_display_destroy.link);
	wl_list_remove(&half_server->wl_display_destroy.link);

	wl_list_init(&half_server->wlr_backend_destroy.link);
	wl_list_init(&half_server->wlr_backend_new_input.link);
	wl_list_init(&half_server->wlr_backend_new_output.link);
	wl_list_remove(&half_server->wlr_backend_destroy.link);
	wl_list_remove(&half_server->wlr_backend_new_input.link);
	wl_list_remove(&half_server->wlr_backend_new_output.link);

	wl_list_init(&half_server->wlr_compositor_destroy.link);
	wl_list_remove(&half_server->wlr_compositor_destroy.link);

	wl_list_init(&half_server->wlr_xdg_shell_destroy.link);
	wl_list_remove(&half_server->wlr_xdg_shell_destroy.link);

	wl_list_init(&half_server->wlr_data_device_manager_destroy.link);
	wl_list_remove(&half_server->wlr_data_device_manager_destroy.link);

	wl_display_destroy_clients(half_server->wl_display);
	wl_display_destroy(half_server->wl_display);
	wlr_backend_destroy(half_server->wlr_backend);
	wlr_compositor_destroy(half_server->wlr_compositor);
	wlr_xdg_shell_destroy(half_server->wlr_xdg_shell);
	wlr_data_device_manager_destroy(half_server->wlr_data_device_manager);

	// destroy all half_seats
	struct half_seat *pos, *tmp;
	wl_list_for_each_safe(pos, tmp, &half_server->seats, link) {
		half_seat_destroy(pos);
	}

	free(half_server);
}

bool half_server_start(struct half_server* half_server) {
	if(!wlr_backend_start(half_server->wlr_backend)) {
		fprintf(stderr, "Failed to start wlr_backend\n");
		return false;
	}

	setenv("WAYLAND_DISPLAY", half_server->socket, true);
	fprintf(stdout, "Started on socket %s\n", half_server->socket);

	// enter event loop
	wl_display_run(half_server->wl_display);

	return true;
}

static void handle_wlr_backend_new_input(
		struct wl_listener* listener,
		void* data) {
	struct half_server* half_server =
			wl_container_of(listener, half_server, wlr_backend_new_input);
	struct wlr_input_device* wlr_input_device = data;

	// create device-specific struct for input-device (if supported)
	if(wlr_input_device->type == WLR_INPUT_DEVICE_KEYBOARD) {
		struct half_keyboard* half_keyboard =
				half_keyboard_create(wlr_input_device);
		// for now just add the keyboard to primary seat, configure in future
		struct half_seat* half_seat =
				wl_container_of(half_server->seats.next, half_seat, link);
		half_seat_add_keyboard(half_seat, half_keyboard);
	} else {
		fprintf(stderr, "Unsupported input device: %s\n", wlr_input_device->name);
	}
}

static void handle_wlr_backend_new_output(
		struct wl_listener* listener,
		void* data) {
	struct half_server* half_server =
			wl_container_of(listener, half_server, wlr_backend_new_output);

	// TODO: try relink with an abandoned half_output for reconnects
	struct half_output* half_output = half_output_create();
	half_output_set_wlr_output(half_output, data);

	wl_list_insert(half_server->outputs.prev, &half_output->link);
}
